//------------------------------------------------------------------------------
// Copyright (C) 2010, Raditex AB
// All rights reserved.
//
// FreeSCADA 
// http://www.FreeSCADA.com
// freescada@freescada.com
//
//------------------------------------------------------------------------------

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <glib.h>

#include <stdio.h>
#include <mbus.h>

int
convert_hex_to_int(char lsb, char msb)
{   
    char buff[3];
    int val;
    
    buff[0] = msb;
    buff[1] = lsb;
    buff[2] = 0;
    
    val = strtol(buff, NULL, 16);

    return val;
}

int
main(int argc, char **argv)
{
    int i = 0, j = 0;
    gchar buff[1024], buff_bin[1024];

    bzero((void *)buff, sizeof(buff));
    bzero((void *)buff_bin, sizeof(buff_bin));

    mbus_frame *frame;
    mbus_frame_data frame_data;

    //
    // 68 1F 1F 68       header of RSP_UD telegram (length 1Fh=31d bytes)
    // 08 02 72          C field = 08 (RSP), address 2, CI field 72H (var.,LSByte first)
    // 78 56 34 12       identification number = 12345678
    // 24 40 01 07       manufacturer ID = 4024h (PAD in EN 61107), generation 1, water
    // 55 00 00 00       TC = 55h = 85d, Status = 00h, Signature = 0000h
    // 03 13 15 31 00    Data block 1: unit 0, storage No 0, no tariff, instantaneous volume,
    //                   12565 l (24 bit integer)
    // DA 02 3B 13 01    Data block 2: unit 0, storage No 5, no tariff, maximum volume flow,
    //                   113 l/h (4 digit BCD)
    // 8B 60 04 37 18 02 Data block 3: unit 1, storage No 0, tariff 2, instantaneous energy,
    // 6.3 Variable Data Structure                   45
    //                      218,37 kWh (6 digit BCD)
    // 18 16                checksum and stopsign
    //

    snprintf(buff, sizeof(buff),
            "681F1F680802727856341224400107550000000313153100DA023B13018B60043718021816");   

    i = 0; j = 0;
    while (i <= strlen(buff))
    {
        buff_bin[j++] = convert_hex_to_int(buff[i++], buff[i++]);
    }

    frame = mbus_frame_new(MBUS_FRAME_TYPE_ANY);
    i = mbus_parse(frame, buff_bin, j);
    //printf("mbus_parse = %d\n", i);
    
    mbus_frame_data_parse(frame, &frame_data);
    //printf("mbus_frame_parse_data = %d\n", i);

    //mbus_frame_print(frame);        
    //mbus_frame_data_print(&frame_data);
    mbus_frame_data_print_xml(&frame_data);
    
    mbus_frame_free(frame);

    return -1;
}

