//------------------------------------------------------------------------------
// Copyright (C) 2010, Raditex AB
// All rights reserved.
//
// FreeSCADA 
// http://www.FreeSCADA.com
// freescada@freescada.com
//
//------------------------------------------------------------------------------

#include "net.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/time.h>
#include <time.h>

#include <glib.h>

//------------------------------------------------------------------------------
// Setup an asyncroniuos
//
// Setup a socket and corresponding GIO Channels and callback functions,
// for asyncronious TCP/IP connections (basically a select/listen server).
//
// XXX: TODO: properly free up resources on error
//
//------------------------------------------------------------------------------
/*
GIOChannel *
fs_net_async_setup(int port, FsNetConnectCallback connect_cb, char *listen_addr)
{
	static const char func_name[] = "fs_net_async_setup";
	GIOChannel *gio_sock;
	int sock, ret, opt = 1;
	struct sockaddr_in s_addr;
	int sockaddr_in_len = sizeof(struct sockaddr_in);

	g_message("%s: port = %d", func_name, port);

	if ((sock = socket(AF_INET,SOCK_STREAM, 0)) < 0)
	{
		g_warning("%s: can't assign fd for socket.", func_name);
		return NULL;
	}

	if ((ret = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt))) == -1)
	{
		g_warning("%s: Failed to set socket option SO_REUSEADDR.", func_name);
	}

	s_addr.sin_port = htons(port);
	s_addr.sin_family = AF_INET;
	if (listen_addr == NULL || strcmp(listen_addr, "0.0.0.0") == 0)
	{
    	g_warning("%s: Falling back to INADDR_ANY.", func_name);
	    s_addr.sin_addr.s_addr = INADDR_ANY;
	}
	else
	{
    	g_warning("%s: Setting listen address to '%s'", func_name, listen_addr);
        
	    //s_addr.sin_addr.s_addr = inet_addr(listen_addr);
        //if (s_addr.sin_addr.s_addr == INADDR_NONE)
        if (inet_aton(listen_addr, &s_addr.sin_addr) == 0)
        {
            g_warning("%s: inet_addr failed, resulted in INADDR_NONE. reverting to INADDR_ANY", func_name);
        }
	    s_addr.sin_addr.s_addr = INADDR_ANY;
    }

	if (bind(sock, (struct sockaddr *) &s_addr, sizeof(s_addr)) < 0)
	{
		g_message("%s: can't bind socket (%d)", func_name, sock);
		return NULL;
   }

   if (getsockname(sock, (struct sockaddr *) &s_addr,
         (socklen_t *) &sockaddr_in_len) < 0)
   {
      g_warning("%s: can't get port number of socket.", func_name);
      return NULL;
   }

   g_message("%s: opened socket as fd (%d) on port (%d) for stream i/o.",
         func_name, sock, ntohs(s_addr.sin_port));

   listen(sock, 1);

	// set up a I/O watch on the socket
	if ((gio_sock = g_io_channel_unix_new(sock)) == NULL)
	{
		g_warning("%s: Couldn't create a new GIOChannel.", func_name);
		return NULL;
	}

	// set up I/O event handlers
	if (!g_io_add_watch(gio_sock, G_IO_IN|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
	    fs_net_async_accept, (gpointer) connect_cb))
	{
		g_warning("%s: Couldn't add watch on GIOChannel.", func_name);
		return NULL;
	}

	g_io_channel_set_encoding(gio_sock, NULL, NULL);
	g_io_channel_set_buffered(gio_sock, FALSE);

	return gio_sock;
}
*/
//------------------------------------------------------------------------------
// Used as a callback function for event on the listening socket. When
// a new connection appears this callback function is called and it
// accepts the socket and creates a new socket for connection.
//
//------------------------------------------------------------------------------
/*
gboolean
fs_net_async_accept(GIOChannel *gio, GIOCondition condition, gpointer data)
{
   static const char func_name[] = "fs_net_async_accept";
   GIOChannel *gio_sock;
   int fd, nsock;
   struct sockaddr_in c_addr;
   int sockaddr_in_len = sizeof(struct sockaddr_in);

   FsNetConnectCallback connect_cb = (FsNetConnectCallback)data;

   fd = g_io_channel_unix_get_fd(gio);

   g_message("%s: new connection (%d).", func_name, fd);

   //nsock = fs_net_socket_accept(fd);
   if ((nsock = accept(fd, (struct sockaddr *) &c_addr,
         (socklen_t *) &sockaddr_in_len)) < 0)
   {
      g_warning("%s: Couldn't accept new connection.", func_name);
      return -1;
   }

   g_message("%s: got new socket (%d)", func_name, nsock);

   if ((gio_sock = connect_cb(nsock, gio, &c_addr)) == NULL)
   {
      g_warning("%s: Connect callback failed to setup a GLIB IO Channel.",
            func_name);
   }

   return TRUE;
}
*/

//------------------------------------------------------------------------------
// Close down a socket.
//------------------------------------------------------------------------------
int
fs_net_socket_close(int sock)
{
	static const char func_name[] = "fs_net_socket_close";
	g_message("%s: close socket: %d", func_name, sock);
	close(sock);
	return -1;
}

//------------------------------------------------------------------------------
// Establish a TCP/IP connection to a server with hostname and port.
// Set up callback functions (given by io_cb) that is supposed to be
// called when data is available from the server, or a socket event
// such as disconnect etc. happens.
//------------------------------------------------------------------------------
GIOChannel *
fs_net_connection_setup(char *hostname, int port, void *io_cb, gpointer cb_data)
{
	static const char func_name[] = "fs_net_connection_setup";
   struct hostent *host_addr;
   struct sockaddr_in s_addr;
   int sock;
   struct timeval time_out;

   g_message("%s: Trying to connect to %s:%d.", func_name, hostname, port);

   GIOChannel *gio_sock;

   if ((sock = socket(AF_INET,SOCK_STREAM, 0)) < 0)
   {
      g_warning("%s: failed to setup a socket.", func_name);
      return NULL;
   }

   s_addr.sin_family = AF_INET;
   s_addr.sin_port = htons(port);

   /* resolve hostname */
   if ((host_addr = gethostbyname(hostname)) == NULL)
   {
      g_warning("%s: unknown host: %s", func_name, hostname);
      return NULL;
   }
   bcopy((char *) (host_addr->h_addr), (char *) (&s_addr.sin_addr),
         host_addr->h_length);

   if (connect(sock, (struct sockaddr *) &s_addr, sizeof(s_addr)) < 0)
   {
      g_warning("%s: Failed to establish connection to %s:%d", func_name,
            hostname, port);
      return NULL;
   }

   // Try to set a timeout... seems to work
   time_out.tv_sec = 5;
   time_out.tv_usec = 0;
   setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, &time_out, sizeof(time_out));
   setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, &time_out, sizeof(time_out));

   /* set up a GLIB IO Channel for the socket */
   if ((gio_sock = g_io_channel_unix_new(sock)) == NULL)
   {
      g_warning("%s: Couldn't create a new GIOChannel.", func_name);
      return NULL;
   }

   /* set up I/O event handler */
   if (!g_io_add_watch(gio_sock, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
         io_cb, cb_data))
   {
      g_warning("%s: Couldn't add watch on GIOChannel.", func_name);
      return NULL;
   }

   //
   // XXX: Why this magic port number ???
   //
   if (port != 2010)
   {
      g_io_channel_set_encoding(gio_sock, NULL, NULL);
      g_io_channel_set_buffered(gio_sock, FALSE);
   }
   else
   {
      g_message("%s: NO encoding and no buffering.", func_name);
   }

   return gio_sock;
}

//------------------------------------------------------------------------------
// Connect an already initialized socket to an GIO channel.
//------------------------------------------------------------------------------
GIOChannel *
fs_net_create_gio(int sock, void *io_cb, gpointer cb_data)
{
    GIOChannel *gio_sock;
   
    /* set up a GLIB IO Channel for the socket */
    if ((gio_sock = g_io_channel_unix_new(sock)) == NULL)
    {
        g_warning("%s: Couldn't create a new GIOChannel.", __PRETTY_FUNCTION__);
        return NULL;
    }

    /* set up I/O event handler */
    if (!g_io_add_watch(gio_sock, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL, io_cb, cb_data))
    {
        g_warning("%s: Couldn't add watch on GIOChannel.", __PRETTY_FUNCTION__);
        return NULL;
    }

    g_io_channel_set_encoding(gio_sock, NULL, NULL);
    g_io_channel_set_buffered(gio_sock, FALSE);
    
    return gio_sock;
}

//------------------------------------------------------------------------------
// Close down a socket and GIO Channel.
//------------------------------------------------------------------------------
gint
fs_net_disconnect(GIOChannel *gio)
{
	static const char func_name[] = "fs_net_disconnect";
	GError *err = NULL;
	int unix_fd;

	if (gio == NULL)
	{
		g_message("%s: gio = NULL no g_io_channel to shut down.", func_name);
		return -1;
	}

	unix_fd = g_io_channel_unix_get_fd(gio);

	g_io_channel_shutdown(gio, TRUE, &err);

	g_message("%s: shutdown g_io_channel.", func_name);
	close(unix_fd);

	return 0;
}

//------------------------------------------------------------------------------
// fs_net_send:
//
// This functions sends a packet on a given GIO Channel:
//------------------------------------------------------------------------------
gint
fs_net_send(GIOChannel *gio, gchar *buff, gint buff_len)
{
   static const char func_name[] = "fs_net_send";
   gint ret;
   gsize len;

//   g_message("%s: [to: %d] packet buff [length = %d]:\n---\n%s\n---\n",
//         func_name, (int) gio, buff_len, buff);
   g_message("%s: [to: %d] packet buff [length = %d]", func_name, (int) gio, buff_len);

   // We need a timeout here...??
   if ((ret = g_io_channel_write(gio, buff, buff_len, &len)) != G_IO_ERROR_NONE)
   {
      g_message("%s: Failed to write data to [%d].", func_name, (int) gio);
      return -1;
   }
   else
   {
      if (len != buff_len)
      {
         g_message("%s: Wrote %d bytes (out of %d requested) to [%d].",
               func_name, len, buff_len, (int) gio);
      }
   }

   return (gint) len;
}

//------------------------------------------------------------------------------
// fs_net_read: 
//
// This function attempts to read one, and only one, packet from the
// IO channel.
//
// It starts by looking for a matching start tag (from a list),
// discarding data preceeding any such matching start tag, and once
// the start tag is found, the data is read and stored until the matching
// end tag is encountered, and then the function returns the read
// data, regardless of whether there is more data available or not.
//
// XXX: Problem. Once a start tag is found the function blocks until the
// matching end tag is found. This is a problem if a slow client starts
// to send a packet but takes long time finish sending it (e.g. a manual
// session by telnet).
//------------------------------------------------------------------------------
gint
fs_net_read(GIOChannel *gio, gchar *buff, gint buff_max_len)
{
	static const char func_name[] = "fs_net_read";
	gint offset = 0;
	GIOStatus ret;
	GError *gerror = NULL;
	gsize bytes_read;

	//printf("DEBUG: Enter fs_net_read: \n");

	// Make sure that the no_tags variable is equal to the actual
	// number of start/end tag pairs:
	gchar *start_tags[] = {"<FSAlarm>", "<FSMessage>", "<FSEvent>", "<FSEventAck>", "<FSSiteStatus>"};
	gchar *end_tags[]   = {"</FSAlarm>", "</FSMessage>", "</FSEvent>", "</FSEventAck>", "</FSSiteStatus>"};
	const gint no_tags = 5;

	gint tag_idx = -1;

	bzero(buff, buff_max_len);

	while (offset < buff_max_len)
	{
		//if ((g_io_channel_get_buffer_condition(gio) & G_IO_IN) == FALSE)
		//{
		//	printf("DEBUG: Predicting blocking IO read.\n");
		//}

		//printf("DEBUG: enter read char: ");
		//fflush(stdout);
		ret = g_io_channel_read_chars(gio, &buff[offset], 1, &bytes_read, &gerror);

		//if (ret == G_IO_STATUS_AGAIN)
		//{
		//	printf("DEBUG: Prevented blocking read, retry...\n");
		//	continue;
		//}

		if (ret != G_IO_STATUS_NORMAL)
		{
         		// This happens when the socket is disconnected. return -1
         		// to indicate to the caller that the socket is shut down.
			g_message("%s: Got a abnormal GIO event.", func_name);
         		return -1;
      		}

      		if (bytes_read == 0)
      		{
         		g_message("%s: ran out of data, premature return [%d].", func_name, offset);
         		break;
      		}

		if (tag_idx == -1)
		{
			//
			// We have not found any start tag yet, so read
			// and discard data until a valid start tag is
			// found.
			//
			gint i, matching_candidates = no_tags;

			for (i = 0; i < no_tags; i++)
			{
				if (strncmp(start_tags[i], buff, offset + 1) == 0)
				{
					if (strlen(start_tags[i]) == offset + 1)
					{
						// Found a valid start tag!
						tag_idx = i;
						g_message("%s: Found start tag '%s'", func_name, start_tags[i]);
						break;
					}
				}
				else
				{
					matching_candidates--;
				}
			}

			if (matching_candidates > 0)
			{
				offset++;
				continue;
			}

         		//
         		// If we reached this point, the data read so far
         		// is incompatible with all tags, so we start over
         		// reading into the buffer at zero offset.
         		
         		g_message("%s: Discarding mismatching tag data '%s'", func_name, buff);
         		
         		//
         		// Return here, since there is more data in the data
         		// stream the read callback function will be called
         		// again, and another fs_net_read call will also
         		// be made.
			return 0;
		}

		//
		// At this point we have found a start tag and
		// the packet is being read into the buffer. We
		// want to read data until we find the matching
		// end tag.
		//
		if (offset > strlen(end_tags[tag_idx]) &&
		    strcmp(end_tags[tag_idx], &buff[offset - strlen(end_tags[tag_idx]) + 1]) == 0)
      		{
			g_message("%s: Found end tag '%s'", func_name, end_tags[tag_idx]);
			break;
		}
		offset++;
	}

	if (tag_idx == -1)
	{
		g_message("%s: Returning because of no matching tag candidates.", func_name);
		return 0;
	}

	g_message("%s: [from: %d] packet buff [length = %d]: \n%s\n---\n",
        	func_name, (int) gio, offset, buff);

	return offset;
}

//------------------------------------------------------------------------------
// fs_net_read_nonblocking: EXPERIMENTAL.
//
// This function is an attempt to implement an nonblocking version of
// the fs_net_read function. The parameters are the same, except for
// the additional offset parameter (int *). The buffer offset is stored in
// this variable instead of a local offset pointer, as it is required 
// to store the buffer offset over several iterations of the nonblocking
// read function.
//
// Caution: It is important that the caller of this function does NOT 
// use the same memory buffer and offset variables for more than one
// GIO channel. 
//
//------------------------------------------------------------------------------
gint
fs_net_read_nonblocking(GIOChannel *gio, FsNetBuffer *data)
{
    static const char func_name[] = "fs_net_read_nonblocking";
    GIOStatus ret;
    GError *gerror = NULL;
    gsize bytes_read;

    // Make sure that the no_tags variable is equal to the actual
    // number of start/end tag pairs:
	gchar *start_tags[] = {"<FSAlarm>", "<FSMessage>", "<FSEvent>", "<FSEventAck>", "<FSSiteStatus>", "<FSProtocolDevice>"};
	gchar *end_tags[]   = {"</FSAlarm>", "</FSMessage>", "</FSEvent>", "</FSEventAck>", "</FSSiteStatus>", "</FSProtocolDevice>"};
	const gint no_tags = 6;

    if (data->offset == 0)
    {
        memset(data->buff, 0, FS_NET_BUFFER_SIZE);
        data->tag_idx = -1; // just to be sure
    }
    else
    {
        g_message("%s: Restart read: offset = %d, looking for end-tag [%s]",
         	      func_name, data->offset, end_tags[data->tag_idx]);
    }

    while (data->offset < FS_NET_BUFFER_SIZE)
    {
        ret = g_io_channel_read_chars(gio, &(data->buff[data->offset]), 1, &bytes_read, &gerror);

        if (ret == G_IO_STATUS_AGAIN)
        {
            g_message("%s: Ran out of non-blocking data. Postpone read until later.", func_name);

            //
            // We use a special return value (-2) to mark that we ran out of 
            // non-blocking data. We return -2 to indicate to the caller that
            // the read was incomplete and should be restarted on the next
            // callback.
            return -2;
        }

        if (ret != G_IO_STATUS_NORMAL)
        {
            // This happens when the socket is disconnected. return -1
            // to indicate to the caller that the socket is shut down.

	        data->offset  = 0;
	        data->tag_idx = -1;
            return -1;
        }

        if (bytes_read == 0)
        {
            g_message("%s: ran out of data, premature return [%d].", func_name, data->offset);
            break;
        }

        if (data->tag_idx == -1)
        {
            //
            // We have not found any start tag yet, so read
            // and discard data until a valid start tag is
            // found.
            //
            gint i, matching_candidates = no_tags;

            for (i = 0; i < no_tags; i++)
            {
                if (strncmp(start_tags[i], data->buff, data->offset + 1) == 0)
                {
                    if (strlen(start_tags[i]) == data->offset + 1)
                    {
                        // Found a valid start tag!
                        data->tag_idx = i;
                        g_message("%s: Found start tag '%s'", func_name,
                                  start_tags[i]);
                        break;
                    }
                }
                else
                {
                    matching_candidates--;
                }
            }

            if (matching_candidates > 0)
            {
                data->offset++;
                continue;
            }

            //
            // If we reached this point, the data read so far
            // is incompatible with all tags, so we start over
            // reading into the buffer at zero offset.
            //
            // AAA: g_message("%s: Discarding mismatching tag data '%s'\n", func_name, buff);
            //
            //memset(buff, 0, offset+1);
            //offset = 0;
            //continue;
            //
            // Return here, since there is more data in the data
            // stream the read callback function will be called
            // again, and another fs_net_read call will also
            // be made.
	        data->offset  = 0;
	        data->tag_idx = -1;
            return 0;
        }

        //
        // At this point we have found a start tag and
        // the packet is being read into the buffer. We
        // want to read data until we find the matching
        // end tag.
        //

        if (data->offset > strlen(end_tags[data->tag_idx])
            && strcmp(end_tags[data->tag_idx],
                     &data->buff[data->offset - strlen(end_tags[data->tag_idx]) + 1]) == 0)
        {
            g_message("%s: Found end string '%s'", func_name, end_tags[data->tag_idx]);
            break;
        }
        data->offset++;
    }

    if (data->tag_idx == -1)
    {
        g_message("%s: Returning because of no matching tag candidates.", func_name);
        data->offset  = 0;
        data->tag_idx = -1;
        return 0;
    }

    g_message("%s: [from: %d] packet buff [length = %d]: \n%s\n---\n",
              func_name, (int) gio, data->offset, data->buff);

    ret = data->offset;
    data->offset  =  0;
    data->tag_idx = -1;
    return ret;
}

//------------------------------------------------------------------------------
// fs_net_read_data_nonblocking: EXPERIMENTAL.
//
//------------------------------------------------------------------------------
gint
fs_net_read_data_nonblocking(GIOChannel *gio, FsNetBuffer *data, gint data_length)
{
    static const char func_name[] = "fs_net_read_data_nonblocking";
    GIOStatus ret;
    GError *gerror = NULL;
    gsize bytes_read;

    if (data->offset == 0)
    {
        memset(data->buff, 0, FS_NET_BUFFER_SIZE);
    }
    else
    {
        g_message("%s: Reading non-blocking data from socket. offset = %d, looking for %d bytes", func_name, data->offset, data_length);
    }

    while (data->offset < FS_NET_BUFFER_SIZE && data->offset < data_length)
    {
        ret = g_io_channel_read_chars(gio, &(data->buff[data->offset]), 1, &bytes_read, &gerror);

        if (ret == G_IO_STATUS_AGAIN)
        {
            g_message("%s: Ran out of non-blocking data [at offset = %d]. Postpone read until later.", func_name, data->offset);

            //
            // We use a special return value (-2) to mark that we ran out of 
            // non-blocking data. We return -2 to indicate to the caller that
            // the read was incomplete and should be restarted on the next
            // callback.
            return -2;
            //break;
        }

        if (ret != G_IO_STATUS_NORMAL)
        {
            // This happens when the socket is disconnected. return -1
            // to indicate to the caller that the socket is shut down.

	        data->offset = 0;
            return -1;
        }

        if (bytes_read == 0)
        {
            g_message("%s: ran out of data, premature return [%d].", func_name, data->offset);
            break;
        }

        //g_message("%s: Read %d bytes: 0x%.2X", func_name, bytes_read, data->buff[data->offset]);

        data->offset++;
 
        if (data->offset == data_length)
        {
            g_message("%s: Reached the targeted number of bytes [%d]", func_name, data_length);
            break;
        }
   }

    g_message("%s: [from: %d] packet buff [length = %d]: ", func_name, (int) gio, data->offset);

    ret = data->offset;
    data->offset  =  0;
    return ret;
}


//------------------------------------------------------------------------------
// Data buffer management:
//
//
static GList *read_buffer_list = NULL;
static gint list_len = 0;
gint
fs_net_buffer_new(GIOChannel *gio)
{
    static gchar *func_name = "fs_net_buffer_new";
    FsNetBuffer *nbdata;

    if ((nbdata = fs_net_buffer_get(gio)) == NULL)
    {
        if ((nbdata = g_malloc(sizeof(FsNetBuffer))) == NULL)
        {
            g_warning("%s: Failed to allocate memory for data buffer.", func_name);
            return -1;
        }

        memset((void *)nbdata, 0, sizeof(FsNetBuffer));
        nbdata->gio     =  gio;
        nbdata->offset  =  0;
        nbdata->tag_idx = -1;

        read_buffer_list = g_list_append(read_buffer_list, (gpointer)nbdata);
        list_len++;

        g_message("%s: A data buffer was registerd for GIO channel %d [%d].",
                  func_name, (gint)gio, list_len);
    }
    else
    {
        g_warning("%s: A data buffer is already associated with GIO channel %d [%d].",
                  func_name, (gint)gio, list_len);
        return -1;
    }

    return 0;
}

void
fs_net_buffer_free(GIOChannel *gio)
{
    static gchar *func_name = "fs_net_buffer_free";
    FsNetBuffer *nbdata;

    if ((nbdata = fs_net_buffer_get(gio)) == NULL)
    {
        g_warning("%s: No data buffer is associated with the GIO channel %d.",
                  func_name, (gint)gio);
        return;        
    }

    read_buffer_list = g_list_remove(read_buffer_list, nbdata);
    g_free(nbdata);
    list_len--;

    g_message("%s: Unregisted buffer for the GIO channel %d [%d].",
              func_name, (gint)gio, list_len);
    return;
}


FsNetBuffer *
fs_net_buffer_get(GIOChannel *gio)
{
    //static gchar *func_name = "fs_net_buffer_get";
    FsNetBuffer *nbdata = NULL;
    GList *item;

    for (item = read_buffer_list; item; item = g_list_next(item))
    {
        nbdata = (FsNetBuffer *)item->data;

        if (nbdata && nbdata->gio == gio)
            return nbdata;
    }

    return NULL;
}


